% Interface to problem-specific part of the encoding:
result(ParentState,Action,ChildState) :- c(ParentState,Action,ChildState,StepCost).
action(ParentState,Action) :- result(ParentState,Action,ChildState).

% GENERAL SEARCH (Section 3.3 in AIMA):

% predicate structure of n/5: n(State,ParentState,Action,PathCost).
%	State: the state in the state space to which the node corresponds
%	ParentState: the state of the node in the search tree that generated this node
%	Action: the action that was applied to the parent to generate the node
%	PathCost: the cost, traditionally denoted by g(n), of the path from the initial state to the node
% (this structure differs a bit from the one in AIMA)
n(State,null,null,0) :- init(State).
n(ChildState,ParentState,Action,PathCost) :- result(ParentState,Action,ChildState), explore(ParentState,Action), g(ParentState,Action,PathCost).
% for uses where only the State of a node is needed, there is also an n/1 predicate:
n(State) :- n(State,ParentState,Action,PathCost).

% INFORMED SEARCH (Section 3.5 in AIMA):
g(State,G) :- n(State,_,_,G).
% h(State,H) is specified in problem-specific part of encoding

% A* SEARCH (Section 3.5.2 in AIMA) in its GRAPH-SEARCH variant (Figure 3.7 in AIMA):
% f(n) = g(n) + h(n):
f(State,GpH) :- g(State,G), h(State,H), GpH=G+H.
% costs are not only needed for nodes (because we create nodes only when they are chosen), but also for node-action pairs:
f(State,Action,GpH) :- g(State,Action,G), h(State,Action,H), GpH=G+H.
g(ParentState,Action,G) :- g(ParentState,ParentG), c(ParentState,Action,_,StepCost), G=ParentG+StepCost.
h(ParentState,Action,H) :- result(ParentState,Action,ChildState), h(ChildState,H).

% The union of the frontier and the explored set is represented by frontier/2 (containing state-action pairs for nodes not yet generated).
frontier_and_explored_states(ParentState,Action) :- n(ParentState), result(ParentState,Action,ChildState), not goal(ParentState).

% The explored set is represented by explored_by_other_step/3, because we need the information from which parent a node has been explored to avoid exploring it again:
explored_by_other_step(ChildState,ParentState,Action) :- n(ChildState,OtherParentState,OtherAction,_), result(ParentState,Action,ChildState), OtherParentState != ParentState.
explored_by_other_step(ChildState,ParentState,Action) :- n(ChildState,OtherParentState,OtherAction,_), result(ParentState,Action,ChildState), OtherAction != Action.

% Contrary to original A*, we may re-insert explored nodes into the frontier. But then these steps must be regarded as suboptimal:
suboptimal_step(State1,Action1) :- frontier_and_explored_states(State1,Action1), result(State1,Action1,ChildState),
                                   explored_by_other_step(ChildState,State1,Action1).
								   
% Replace node in frontier when a cheaper path to that node has been found:
suboptimal_step(State1,Action1) :- frontier_and_explored_states(State1,Action1), result(State1,Action1,ChildState),
                                   frontier_and_explored_states(State2,Action2), result(State2,Action2,ChildState),
                                   f(State1,Action1,F1), f(State2,Action2,F2), F1 > F2.


% expand nodes until a goal state is reached:
goal_found(State) :- n(State), goal(State).
goal_found :- goal_found(State).

% do not accept found goal if node with lower f is on the frontier (to ensure optimality of found solutions):
:- goal_found(State), explore(ParentState,Action), result(ParentState,Action,State), f(State,F),
   frontier_and_explored_states(OtherState,OtherAction), not explore(OtherState,OtherAction), not suboptimal_step(OtherState,OtherAction), f(OtherState,OtherAction,OtherF), OtherF < F.

% choose a node from the frontier:
{ explore(State,Action) } :- frontier_and_explored_states(State,Action), not suboptimal_step(State,Action).

% return failure if the frontier is empty and no goal has been found:
frontier_nonempty :- frontier_and_explored_states(State,Action), not suboptimal_step(State,Action), not explore(State,Action).
failure :- not frontier_nonempty, not goal_found.
:- not failure, not goal_found.

% main heuristic statement encoding A*:
#heuristic explore(ParentState,Action) : frontier_and_explored_states(ParentState,Action), f(ParentState,Action,PathCost). [-PathCost@3]
% close unassigned choices when goal is found:
#heuristic F explore(ParentState,Action) : frontier_and_explored_states(ParentState,Action), goal_found. [1@2]

% compute path to goal and corresponding cost:
path_to_goal(ParentState,Action,ChildState) :- explore(ParentState,Action), result(ParentState,Action,ChildState), goal_found(ChildState).
path_to_goal(ParentState,Action,ChildState) :- explore(ParentState,Action), result(ParentState,Action,ChildState), path_to_goal(ChildState,_,_).
cost_to_goal(G) :- goal_found(GoalState), g(GoalState,G).

% predicates to be printed (Alpha command-line: "--filter path_to_goal --filter cost_to_goal --filter failure")
% #show path_to_goal/3.
% #show cost_to_goal/1.
% #show failure/0.
